#version 430 core

out vec4 final_color;

in VertexData {
	vec3 position;
	vec3 normal;
	vec3 color;
} vert;

uniform vec3 camPos;

layout(location=3) uniform vec3 albedo;
layout(location=4) uniform float metallic;
layout(location=5) uniform float roughness;
layout(location=6) uniform float ao;

struct PointLight {
	vec3 color;
	vec3 pos;
    vec3 attenuation;
};

struct DirLight {
	vec3 color;
	vec3 dir;
};

uniform int dirLightLength;
uniform DirLight dirLight[10];
uniform int pointLightLength;
uniform PointLight pointLight[10];

const float PI = 3.14159265359;

vec3 fresnelSchlick(float cosTheta, vec3 f0){
	// how much light is reflected:
	return f0+(1.0-f0) * pow(1-cosTheta, 5.0);
}

float distributionGGX(vec3 n, vec3 h, float roughness) {
	float a = roughness*roughness;
	float a2 = a*a;
	float ndoth=max(dot(n,h),0.0);
	float ndoth2=ndoth*ndoth;

	float denom = (ndoth2 * (a2-1.0) +1.0);
	denom=PI*denom*denom;

	return a2/denom;
}

float geometrySchlickGGX(float ndotv, float roughness) {
	float r = (roughness+1.0);
	float k = (r*r) / 8.0;

	float denom=ndotv*(1.0-k)+k;
	return ndotv/denom;
}

float geometrySmith(vec3 n, vec3 v, vec3 l, float roughness){
	float ndotv=max(dot(n,v),0.0);
	float ndotl=max(dot(n,l),0.0);
	float ggx2=geometrySchlickGGX(ndotv,roughness);
	float ggx1=geometrySchlickGGX(ndotl,roughness);

	return ggx1*ggx2;
}

void main() {
	vec3 normal=normalize(vert.normal);
	vec3 toEye=normalize(camPos-vert.position);

	vec3 lo=vec3(0.0);

	vec3 f0 = vec3(0.04);
	f0=mix(f0,albedo,metallic);

	for(int i=0;i<pointLightLength;i++) {
		vec3 l=normalize(pointLight[i].pos-vert.position);
		vec3 h=normalize(toEye+l);

		float distance = length(pointLight[i].pos-vert.position);
		float attenuation = 1.0 /(distance*distance);
		vec3 radiance = pointLight[i].color * attenuation;

		vec3 f= fresnelSchlick(max(dot(h,toEye),0.0), f0);

		float ndf=distributionGGX(normal,h,roughness);
		float g=geometrySmith(normal,toEye,l,roughness);

		vec3 numerator=ndf*g*f;
		float denominator=4.0*max(dot(normal,toEye),0.0)*max(dot(normal,l),0.0);
		vec3 specular = numerator/max(denominator,0.001);

		vec3 kS = f;
		vec3 kD = vec3(1.0)-kS;

		kD *=1.0-metallic;
		
		float ndotl=max(dot(normal,l),0.0);
		lo+=(kD*albedo/PI+specular)*radiance*ndotl;
	}
	for(int i=0;i<dirLightLength;i++) {
		vec3 l=normalize(dirLight[i].dir);
		vec3 h=normalize(toEye+l);

		vec3 radiance = dirLight[i].color;

		vec3 f= fresnelSchlick(max(dot(h,toEye),0.0), f0);

		float ndf=distributionGGX(normal,h,roughness);
		float g=geometrySmith(normal,toEye,l,roughness);

		vec3 numerator=ndf*g*f;
		float denominator=4.0*max(dot(normal,toEye),0.0)*max(dot(normal,l),0.0);
		vec3 specular = numerator/max(denominator,0.001);

		vec3 kS = f;
		vec3 kD = vec3(1.0)-kS;

		kD *=1.0-metallic;
		
		float ndotl=max(dot(normal,l),0.0);
		lo+=(kD*albedo/PI+specular)*radiance*ndotl;
	}

	vec3 ambient = 0.03*vert.color*albedo*ao;
	vec3 color = ambient+lo;

	color=color/(color+vec3(1.0));
	color=pow(color,vec3(1.0/2.2));
	final_color=vec4(color,1.0);
}


