#version 430 core

// Structs
struct Material
{
	sampler2D diffuseSampler;
	float ambient;
	float diffuse;
	float specular;
	float alpha;
};

struct AmbientLight
{
	vec3 color;
};

struct DirLight
{
	vec3 dir;
	// vec3 color;
};

// Inputs
in VS_OUT
{
    vec3 fWorldPos;
    vec3 fNorm;
    vec2 fTexCoords;
} vs_in;

// Outputs
out vec4 result;

// Uniforms
uniform Material material;
uniform AmbientLight ambientLight;
uniform DirLight dirLight;
uniform vec3 viewPos;

uniform sampler2D shadowTexture;
uniform sampler2D edgeTexture;
uniform mat4 shadowViewProjectionMatrix;
uniform float shadowNear;
uniform float shadowFar;

// number of levels
const int lightLevels = 20;
const float lightScaleFactor = 1.0 / lightLevels;

const int colorLevels = 5;
const float colorScaleFactor = 1.0 /colorLevels;

// Logic
float calcDiffuse(vec3 normal, vec3 lightDir) {
    return max(0.0, dot(normal, -lightDir)) * material.diffuse;
}

float calcSpecular(vec3 normal, vec3 viewDir, vec3 lightDir) {
	vec3 ref = reflect(lightDir, normal);
	return pow(max(0.0, dot(ref, -viewDir)), material.alpha) * material.specular;
}

vec3 calcAmbientLight(vec3 diffColor) {
	return ambientLight.color * diffColor * material.ambient;
}

vec3 calcDirLight(vec3 diffColor, vec3 normal, vec3 viewDir)
{
	float diffuse = calcDiffuse(normal, dirLight.dir);
	vec3 diffuseColor = diffColor * floor(diffuse * lightLevels) * lightScaleFactor;

	float specular = 0.0;

	if(dot(dirLight.dir, normal) > 0.0)
	{
		specular = calcSpecular(normal, viewDir, dirLight.dir);
	}

	// limit specular
	vec3 H = normalize(viewDir+dirLight.dir);
	float specMask = (pow(dot(H, normal), material.alpha) > 0.4) ? 1 : 0;

	return (diffuseColor + specular * specMask);
}

// All components are in the range [01], including hue.
vec3 rgb2hsv(vec3 c)
{
    vec4 K = vec4(0.0, -1.0 / 3.0, 2.0 / 3.0, -1.0);
    vec4 p = mix(vec4(c.bg, K.wz), vec4(c.gb, K.xy), step(c.b, c.g));
    vec4 q = mix(vec4(p.xyw, c.r), vec4(c.r, p.yzx), step(p.x, c.r));

    float d = q.x - min(q.w, q.y);
    float e = 1.0e-10;
    return vec3(abs(q.z + (q.w - q.y) / (6.0 * d + e)), d / (q.x + e), q.x);
}
 

// All components are in the range [01], including hue.
vec3 hsv2rgb(vec3 c)
{
    vec4 K = vec4(1.0, 2.0 / 3.0, 1.0 / 3.0, 3.0);
    vec3 p = abs(fract(c.xxx + K.xyz) * 6.0 - K.www);
    return c.z * mix(K.xxx, clamp(p - K.xxx, 0.0, 1.0), c.y);
}

float linearizeDepth (float depth) {
	float z = depth * 2.0 - 1.0; // Back to NDC
	return (2.0 * shadowNear * shadowFar) / (shadowFar + shadowNear - z * (shadowFar - shadowNear));
}

float calculateShadow(vec3 normal) 
{
	vec4 posInShadowProj = shadowViewProjectionMatrix * vec4(vs_in.fWorldPos, 1.0);
	posInShadowProj = posInShadowProj / posInShadowProj.w; // Homogenization
	posInShadowProj.xyz = (posInShadowProj.xyz + vec3(1.0)) / 2.0;
	float currentDepth =  posInShadowProj.z;

	float bias = max(0.001, 0.020 * (1.0 - dot(normal, -dirLight.dir)));
	bias /= (shadowFar - shadowNear);

	vec2 texelSize = 1.0 / textureSize(shadowTexture, 0);
	float shadow = 0;
	for (int x = -2; x <= 2; ++x) {
		for (int y = -2; y <= 2; ++y) {
			float pcfDepth = texture(shadowTexture, posInShadowProj.xy + vec2(x, y) * texelSize).r;
			shadow += currentDepth - bias >= pcfDepth ? 1.0 : 0.0;
		}
	}
	shadow /= 25.0;

	if (currentDepth > 1.0) {
		shadow = 0.0;
	}
	
	return shadow;
}

void main()
{
	vec3 normal = normalize(vs_in.fNorm);
	vec3 viewDir = normalize(vs_in.fWorldPos - viewPos);
	vec3 diffColor = vec3(texture(material.diffuseSampler, vs_in.fTexCoords));

	// Check if edge
	float edge = texture(edgeTexture, gl_FragCoord.xy / textureSize(edgeTexture, 0)).r;
	if (edge > 0.5) {
		result = vec4(0.0, 0.0, 0.0, 1.0);
		return;
	}

	// Toon shader
	vec3 hsvColor = rgb2hsv(diffColor);
	hsvColor = vec3(hsvColor.x, hsvColor.y, floor(hsvColor.z * colorLevels) * colorScaleFactor);
	diffColor = hsv2rgb(hsvColor);

	// Calculate color:
	vec3 sum = calcAmbientLight(diffColor);
	sum += calcDirLight(diffColor, normal, viewDir) * (1.0 - calculateShadow(normal));

    result = vec4(sum, 1.0);
}
